import numpy as np

from . import jtor

def test_psinorm_range():
    """Test that the profiles produce finite values outside core"""

    for profiles in [jtor.ConstrainPaxisIp(1e3, 2e5, 2.0),
                     jtor.ConstrainBetapIp(1.0, 2e5, 2.0)]:

        # Need to give a plasma psi
        R, Z = np.meshgrid(np.linspace(0.5,1.5,33), np.linspace(-1,1,33), indexing='ij')
        psi = np.exp((-(R - 1.0)**2 - Z**2)*3) + np.exp((-(R - 1.0)**2 - (Z + 1)**2)*3)

        current_density = profiles.Jtor(R, Z, psi)
        assert np.all(np.isfinite(current_density))
        
        assert profiles.pprime(1.0) == 0.0
        assert profiles.pprime(1.1) == 0.0
        assert np.isfinite(profiles.pprime(-0.32))
        
        assert profiles.ffprime(1.0) == 0.0
        assert profiles.ffprime(1.1) == 0.0
        assert np.isfinite(profiles.ffprime(-0.32))



"""
Tests of the linear solver
"""

from . import multigrid

import numpy as np

###### Laplacian in 2D

def test_direct_laplacian():
    nx = 65
    ny = 65

    Lx = 2.
    Ly = 3.
    
    dx = Lx/(nx - 1)
    dy = Ly/(ny - 1)

    xx, yy = np.meshgrid(np.linspace(0,Lx,nx), np.linspace(0,Ly,ny))

    solution = np.exp( - ( (xx - 0.5)**2 + (yy - 0.5)**2 ) / 0.4**2 )
    
    A = multigrid.LaplacianOp()
    rhs = A(solution, dx, dy)

    # Copy boundaries
    rhs[:,0] = solution[:,0]
    rhs[0,:] = solution[0,:]
    rhs[:,-1] = solution[:,-1]
    rhs[-1,:] = solution[-1,:]
    
    solve = multigrid.MGDirect(multigrid.LaplaceSparse(Lx, Ly)(nx, ny))
    x = solve(None, rhs)
    
    assert np.allclose(x, solution)


def test_multigrid_laplacian():
    nx = 65
    ny = 65

    Lx = 2.
    Ly = 3.
    
    dx = Lx/(nx - 1)
    dy = Ly/(ny - 1)

    xx, yy = np.meshgrid(np.linspace(0,Lx,nx), np.linspace(0,Ly,ny))

    solution = np.exp( - ( (xx - 0.5)**2 + (yy - 0.5)**2 ) / 0.4**2 )
    
    A = multigrid.LaplacianOp()
    rhs = A(solution, dx, dy)

    # Copy boundaries
    rhs[:,0] = solution[:,0]
    rhs[0,:] = solution[0,:]
    rhs[:,-1] = solution[:,-1]
    rhs[-1,:] = solution[-1,:]
    
    solve = multigrid.createVcycle(nx,ny, multigrid.LaplaceSparse(Lx, Ly),
                                   ncycle=50, niter=50, nlevels=4, direct=True)

    xinput = np.zeros(rhs.shape)
    xinput[:,0] = solution[:,0]
    xinput[0,:] = solution[0,:]
    xinput[:,-1] = solution[:,-1]
    xinput[-1,:] = solution[-1,:]
    x = solve(xinput, rhs)
    
    assert np.allclose(x, solution, atol=1e-6)



import numpy as np

from . import quadrature

def test_unity():
    # Sum of weights is one i.e quadrature works out average over area

    for n in [1,3,6]:
        ps = quadrature.polygon_quad([(0,0), (0,1), (1,1), (1,0)], n=n)
        assert len(ps) == 2 * n # Two triangles
        assert np.isclose(sum(weight for r,z,weight in ps), 1.0)

def test_integral():
    # Limits of integration
    x0,x1 = 0.4, 1.3
    y0,y1 = -0.2, 3.1

    def func(x,y):
        return x**2 - y**3 + x*y

    exact_integral = (x1**3 - x0**3)*(y1 - y0)/3 - (y1**4 - y0**4)*(x1 - x0)/4 + (x1**2 - x0**2)*(y1**2 - y0**2)/4

    # A 1st order method can't integrate this polynomial exactly
    quad1 = quadrature.polygon_quad([(x0,y0), (x0,y1), (x1,y1), (x1,y0)], n=1) 
    assert not np.isclose(quadrature.average(func, quad1), exact_integral / ((x1-x0)*(y1-y0)))
    
    for n in [3,6]:
        # Higher order methods can
        quad1 = quadrature.polygon_quad([(x0,y0), (x0,y1), (x1,y1), (x1,y0)], n=n)
        assert len(quad1) == 2*n
        assert np.isclose(quadrature.average(func, quad1), exact_integral / ((x1-x0)*(y1-y0)))




from . import _fileutils

try:
    from io import StringIO
except:
    # Python 2
    from StringIO import StringIO

def test_f2s():
    assert _fileutils.f2s(0.0)  == " 0.000000000E+00"
    assert _fileutils.f2s(1234) == " 1.234000000E+03"
    assert _fileutils.f2s(-1.65281e12) == "-1.652810000E+12"
    assert _fileutils.f2s(-1.65281e-2) == "-1.652810000E-02"

    
    
def test_ChunkOutput():
    output = StringIO()
    co = _fileutils.ChunkOutput(output)

    for val in [1.0, -3.2, 6.2e5, 8.7654e-12, 42., -76]:
        co.write(val)
    
    assert output.getvalue() == """ 1.000000000E+00-3.200000000E+00 6.200000000E+05 8.765400000E-12 4.200000000E+01
   -76"""
    



import numpy

from io import StringIO

from . import _geqdsk

def test_writeread():
    """
    Test that data can be written then read back
    """
    nx = 65
    ny = 65
    
    # Create a dummy dataset
    data = {"nx":nx, "ny":ny,
            "rdim":2.0,
            "zdim":1.5,
            "rcentr":1.2,
            "bcentr":2.42,
            "rleft": 0.5,
            "zmid":0.1,
            "rmagx": 1.1,
            "zmagx": 0.2,
            "simagx": -2.3,
            "sibdry": 0.21,
            "cpasma": 1234521,
            "fpol":numpy.random.rand(nx),
            "pres":numpy.random.rand(nx),
            "qpsi":numpy.random.rand(nx),
            "psi":numpy.random.rand(nx,ny)
    }
    
    output = StringIO()

    # Write to string
    _geqdsk.write(data, output)
    
    # Move to the beginning of the buffer
    output.seek(0)
        
    # Read from string
    data2 = _geqdsk.read(output)
    
    # Check that data and data2 are the same
    for key in data:
        numpy.testing.assert_allclose(data2[key], data[key])
        



from . import polygons
import numpy as np

def test_nointersect():
    assert polygons.intersect([0,1], [0,0], [0,1], [1,1]) == False
    assert polygons.intersect([0,1], [0,0], [2,2], [1,-1]) == False

# Two line segments
def test_lineintersect():
    assert polygons.intersect([0,1], [0,0], [0.5,0.5], [1,-1]) == True

# Two squares
def test_squareintersect():
    assert polygons.intersect([0,1,1,0], [0,0,1,1],
                              [-0.5, 0.5, 0.5, -0.5], [0.5, 0.5, 1.5, 1.5]) == True
    
def test_area():
    assert np.isclose(polygons.area([(0,0), (0,1), (1,1), (1,0)]), 1.0)
    assert np.isclose(polygons.area([(0,0), (0,1), (1,0)]), 0.5)

###### clockwise

def test_clockwise():
    assert not polygons.clockwise([(0,0),(1,0),(1,1),(0,1)])
    assert  polygons.clockwise([(0,0),(0,1),(1,1),(1,0)])


###### triangulate
    
def test_triangulate_three():
    # Always returns a triangle in clockwise order
    assert polygons.triangulate([(0,0), (0,1), (1,0)]) == [[(0,0), (0,1), (1,0)]]
    assert polygons.clockwise(polygons.triangulate([(0,0), (1,0), (0,1)])[0])

def test_triangulate_four():
    square = [(0,0), (0,1), (1,1), (1,0)]
    result = polygons.triangulate(square)
    assert len(result) == 2 # Two triangles
    assert len(square) == 4 # Original not modified
    # Both triangles have positive area
    assert polygons.area(result[0]) > 0.0
    assert polygons.area(result[1]) > 0.0
    # Sum of areas equal to original area
    assert np.isclose(polygons.area(square), polygons.area(result[0]) + polygons.area(result[1]))

    



from .shaped_coil import ShapedCoil

import numpy as np

def test_area():
    coil1 = ShapedCoil([(0.95, 0.1), (0.95, 0.2), (1.05, 0.2), (1.05, 0.1)], current = 100.)
    assert np.isclose(coil1.area, 0.01)

def test_move_R():
    """
    Changing major radius property R changes all filament locations
    """

    dR = 0.6
    coil1 = ShapedCoil([(0.95, 0.1), (0.95, 0.2), (1.05, 0.2), (1.05, 0.1)], current = 100.)
    coil2 = ShapedCoil([(0.95+dR, 0.1), (0.95+dR, 0.2), (1.05+dR, 0.2), (1.05+dR, 0.1)], current = 100.)

    # Shift coil1 to same location as coil2
    coil1.R += dR
    
    assert np.isclose(coil1.controlPsi(0.4, 0.5),
                      coil2.controlPsi(0.4, 0.5))
    
    assert np.isclose(coil1.controlBr(0.3, -0.2),
                      coil2.controlBr(0.3, -0.2))

    assert np.isclose(coil1.controlBz(1.75, 1.2),
                      coil2.controlBz(1.75, 1.2))

def test_move_Z():
    """
    Changing height property Z changes all filament locations
    """

    dZ = 0.4
    coil1 = ShapedCoil([(0.95, 0.1), (0.95, 0.2), (1.05, 0.2), (1.05, 0.1)], current = 100.)
    coil2 = ShapedCoil([(0.95, 0.1 + dZ), (0.95, 0.2 + dZ), (1.05, 0.2 + dZ), (1.05, 0.1 + dZ)], current = 100.)

    # Shift coil1 to same location as coil2
    coil1.Z += dZ
    
    assert np.isclose(coil1.controlPsi(0.4, 0.5),
                      coil2.controlPsi(0.4, 0.5))
    
    assert np.isclose(coil1.controlBr(0.3, -0.2),
                      coil2.controlBr(0.3, -0.2))

    assert np.isclose(coil1.controlBz(1.75, 1.2),
                      coil2.controlBz(1.75, 1.2))



"""
Quadrature rules for averaging over polygons

Note: integration weights set so that sum of weights is 1 and gives
the average of a function over the polygon, not the integral.

License
-------

Copyright 2019 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.
"""

from . import polygons

def triangle_quad(triangle, n = 6):
    """
    Given a triangle in the form [(r1,z1), (r2,z2), (r3,z3)]
    returns a list of evaluation points [(r,z,weight),...]

    n   number of quadrature points.
        currently: 1, 3 or 6

    Coefficients taken from http://www.cs.rpi.edu/~flaherje/pdf/fea6.pdf
    Joseph E. Flaherty course notes, Rensselaer Polytechnic Institute
    """

    assert len(triangle) == 3
    
    r1, z1 = triangle[0]
    r2, z2 = triangle[1]
    r3, z3 = triangle[2]
    
    if n == 1:
        # One point in the middle of the triangle
        return [((r1 + r2 + r3)/3, (z1 + z2 + z3)/3, 1.0)]
    
    elif n == 3:
        return [((4*r1 + r2 + r3)/6, (4*z1 + z2 + z3)/6, 1./3),
                ((r1 + 4*r2 + r3)/6, (z1 + 4*z2 + z3)/6, 1./3),
                ((r1 + r2 + 4*r3)/6, (z1 + z2 + 4*z3)/6, 1./3)]

    elif n == 6:
        a = 0.816847572980459
        b = 0.5*(1. - a)
        
        c = 0.108103018168070
        d = 0.5*(1. - c)
        
        return [((a*r1 + b*r2 + b*r3), (a*z1 + b*z2 + b*z3), 0.109951743655322),
                ((b*r1 + a*r2 + b*r3), (b*z1 + a*z2 + b*z3), 0.109951743655322),
                ((b*r1 + b*r2 + a*r3), (b*z1 + b*z2 + a*z3), 0.109951743655322),
                ((c*r1 + d*r2 + d*r3), (c*z1 + d*z2 + d*z3), 0.223381589678011),
                ((d*r1 + c*r2 + d*r3), (d*z1 + c*z2 + d*z3), 0.223381589678011),
                ((d*r1 + d*r2 + c*r3), (d*z1 + d*z2 + c*z3), 0.223381589678011)]
    else:
        raise ValueError("Quadrature not available for n={}".format(n))
    
            
def polygon_quad(polygon, n = 6):
    """
    Given a polygon in the form [(r1,z1), (r2,z2), (r3,z3), ...]
    calculates a set of quadrature points and weights, by splitting
    the polygon into triangles. 
    
    returns a list of evaluation points and weights [(r,z,weight),...]
    These are normalised to calculate the average value of a function
    over the polygon; multiply by the area to get the integral.

    n   number of quadrature points in each triangle
        currently: 1, 3 or 6
    """

    # Split polygon into triangles
    triangles = polygons.triangulate(polygon)
    
    # Calculate the area of each triangle, to get a relative weighting
    areas = [polygons.area(triangle) for triangle in triangles]
    total_area = sum(areas)

    quadrature = [] # List of all points
    for triangle, area in zip(triangles, areas):
        points = triangle_quad(triangle, n = n) # Quadrature points for this triangle
        quadrature += [(r, z, w * area / total_area) for r, z, w in points] # Modify the weights

    return quadrature

def average(func, quad):
    """
    Average func(r,z) using given quadrature 
    """
    return sum(func(r,z) * w for r,z,w in quad)
